/*
 * Copyright (c) 2008 Lukas Mejdrech
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup net
 * @{
 */

/** @file
 */

#include <async.h>
#include <errno.h>
#include <stdio.h>
#include <ipc/ipc.h>
#include <ipc/services.h>
//#include <sys/mman.h>

#include "../err.h"
#include "../measured_strings.h"
#include "../messages.h"
#include "../modules.h"

#include "netif.h"

#define DEFAULT_MTU	1500

#define NAME	"lo - loopback interface"

netif_globals_t	netif_globals;

void	change_state( netif_device_ref device, netif_state_t state );
int	change_state_message( netif_device_id_t device_id, netif_state_t state );
int	netif_create( netif_device_id_t device_id, netif_device_ref * device );
int	netif_call( ipc_callid_t callid );
int	netif_initialize( void );
void	netif_print_name( void );
int	netif_probe_auto_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 );
int	netif_probe_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 );
int	netif_send_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 );
int	netif_start_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 );
int	netif_stop_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 );

void change_state( netif_device_ref device, netif_state_t state ){
	device->state = state;
	ll_message( device, NET_LL_DEVICE_STATE_CHANGED, device->state, NULL, NULL, NULL, NULL );
}

int change_state_message( netif_device_id_t device_id, netif_state_t state ){
	ERROR_DECLARE;

	netif_device_ref	device;

	ERROR_PROPAGATE( netif_device_find( device_id, & device ));
	change_state( device, state );
	return EOK;
}

int netif_create( netif_device_id_t device_id, netif_device_ref * device ){
	ERROR_DECLARE;

	if( netif_device_map_count( & netif_globals.netif_device_map ) > 0 ){
		return EXDEV;
	}else{
		* device = ( netif_device_ref ) malloc( sizeof( netif_device_t ));
		if( !( * device )) return ENOMEM;
		( ** device ).device_id = device_id;
		( ** device ).ll_registered = -1;
		( ** device ).specific = NULL;
		netif_device_stats_null( &(( ** device ).stats ));
		( ** device ).state = NETIF_STOPPED;
		( ** device ).flags = NULL;
		( ** device ).mtu = DEFAULT_MTU;
		if( ERROR_OCCURED( netif_device_map_add( & netif_globals.netif_device_map, ( ** device ).device_id, * device ))){
			free( * device );
			* device = NULL;
			return ERROR_CODE;
		}
	}
	return EOK;
}

int netif_call( ipc_callid_t callid ){
	return EOK;
}

int netif_initialize( void ){
	ipcarg_t	phonehash;

	return REGISTER_ME( SERVICE_LO, & phonehash );
}

void netif_print_name( void ){
	printf( NAME );
}

int netif_probe_auto_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 ){
/*	ERROR_DECLARE;

	netif_device_ref	device;

	ERROR_PROPAGATE( netif_create( arg1, & device ));
	ipc_call_sync_3_3( netif_globals.networking_phone, NET_NETWORKING_DEVICE, device->device_id, NULL, NULL, NULL, NULL, NULL );
*/	return ENOTSUP;
}

int netif_probe_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 ){
	ERROR_DECLARE;

	netif_device_ref	device;
	aid_t			message;
	ipc_call_t		answer;
	measured_string_t	configuration[ 1 ] = {{ "MTU", 3 }};
	int			count = 1;
	measured_string_ref	settings;
	char *			data;

	// create a new device
	ERROR_PROPAGATE( netif_create( arg1, & device ));
	// get configuration
	message = async_send_2( netif_globals.networking_phone, NET_NETWORKING_GET_DEVICE_CONFIGURATION, device->device_id, count, & answer );
	// send names and get settings
	if( ERROR_OCCURED( measured_strings_send( netif_globals.networking_phone, configuration, count ))
	|| ERROR_OCCURED( measured_strings_return( netif_globals.networking_phone, & settings, & data, count ))){
		async_wait_for( message, NULL );
		return ERROR_CODE;
	}
	// MTU is the first one
	if( settings && ( settings[ 0 ].value )){
		device->mtu = strtoul( settings[ 0 ].value, NULL, 0 );
	}else{
		device->mtu = DEFAULT_MTU;
	}
	// print the settings
	printf("\n -MTU=%d", device->mtu );
	free( settings );
	free( data );
	// end request
	async_wait_for( message, NULL );
	return EOK;
}

int netif_send_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 ){
	ERROR_DECLARE;

	netif_device_ref	device;

	ERROR_PROPAGATE( netif_device_find( arg1, & device ));
	if( device->state == NETIF_ACTIVE ){
		++ device->stats.tx_packets;
		++ device->stats.rx_packets;
		// TODO packet size
		//device->stats->tx_bytes += ;
		//device->stats->rx_bytes += ;
		ll_message( device, NET_LL_RECEIVED, arg2, NULL, NULL, NULL, NULL );
		return EOK;
	}else{
		return EPERM;
	}
}

int netif_start_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 ){
	return change_state_message( arg1, NETIF_ACTIVE );
}

int netif_stop_message( ipcarg_t arg1, ipcarg_t arg2, ipcarg_t arg3, ipcarg_t * result1, ipcarg_t * result2, ipcarg_t * result3 ){
	return change_state_message( arg1, NETIF_STOPPED );
}

/** @}
 */
